import {
    IHttp,
    IHttpRequest,
    IHttpResponse,
    ILogger,
    IRead,
} from "@rocket.chat/apps-engine/definition/accessors";
import { EmptyResponseError } from "../errors/EmptyResponseError";
import { JsonParseError } from "../errors/JsonParseError";
import { RcSdkRole } from "./RcSdkRole";
import { RcSdkTeam } from "./RcSdkTeam";
import { RcSdkUser } from "./RcSdkUser";

type RequestCallback = (
    url: string,
    options: IHttpRequest,
) => Promise<IHttpResponse>;

export class RcSdk {
    public user: RcSdkUser;
    public team: RcSdkTeam;
    public role: RcSdkRole;
    public readonly userId: string;
    private readonly baseUrl = "http://localhost:3000/api/v1";
    private readonly http: IHttp;
    private readonly logger: ILogger;
    private readonly headers: any;

    constructor(
        username: string,
        password: string,
        http: IHttp,
        logger: ILogger,
    ) {
        this.userId = username;
        this.headers = {
            "X-User-Id": username,
            "X-Auth-Token": password,
            "Content-Type": "application/json",
        };
        this.http = http;
        this.logger = logger;
        this.user = new RcSdkUser(this);
        this.team = new RcSdkTeam(this);
        this.role = new RcSdkRole(this);
    }

    public get(url: string, content?: any): Promise<any> {
        return this.request(this.http.get, url, content);
    }

    public post(url: string, content?: any): Promise<any> {
        return this.request(this.http.post, url, content);
    }

    public put(url: string, content?: any): Promise<any> {
        return this.request(this.http.put, url, content);
    }

    public del(url: string, content?: any): Promise<any> {
        return this.request(this.http.del, url, content);
    }

    public patch(url: string, content?: any): Promise<any> {
        return this.request(this.http.patch, url, content);
    }

    public parseResponse(response: IHttpResponse): any {
        if (!response.content) {
            throw new EmptyResponseError();
        }
        let content: any;
        try {
            content = JSON.parse(response.content);
        } catch (e) {
            throw new JsonParseError();
        }
        return content;
    }

    private buildUrl(url: string): string {
        return `${this.baseUrl}/${url}`;
    }

    private async buildOptions(content?: any): Promise<IHttpRequest> {
        const options: IHttpRequest = {
            headers: this.headers,
        };
        if (content !== undefined) {
            options.content = JSON.stringify(content);
        }
        return options;
    }

    private async request(
        method: RequestCallback,
        url: string,
        content?: any,
    ): Promise<IHttpResponse> {
        this.logger.debug(`API Request ${url}`, content);
        const response = await method.bind(this.http)(
            this.buildUrl(url),
            await this.buildOptions(content),
        );
        const o = this.parseResponse(response);
        this.logger.debug(`API Response`, o);
        return o;
    }
}
